// Copyright 2021 CoreOS, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

use crate::cmdline::*;
use anyhow::{bail, Result};

use libcoreinst::blockdev::*;

pub fn verify_unique_fs(config: VerifyUniqueFsLabelConfig) -> Result<()> {
    let pts = get_filesystems_with_label(&config.label, config.rereadpt)?;
    let count = pts.len();
    if count != 1 {
        bail!(
            "System has {} devices with a filesystem labeled '{}': {:?}",
            count,
            config.label,
            pts
        );
    }
    Ok(())
}
